#!/usr/bin/env python

import operator
from collections import defaultdict
import itertools
import logging
import math
from util import *


class Predictor(object):
    def __init__(self, data, options):
    	self.options = options
        self.data = data
        self.precompute()
    
    def precompute(self):
        """
        Given some training data, precompute anything that would be useful for making
        predictions later.
        
        does not return anything
        """
        pass

    def predict(self, user_id, movie_id):
        """
        Return a prediction for how much this user would like this movie, as a float
        between 1.0 and 5.0
        """
        return 3.0

class AverageBased(Predictor):
    """A silly predictor that always predicts the average rating."""
    def precompute(self):
        n = float(len(self.data))
        self.avg = mean(t[2] for t in self.data)
    
    def predict(self, user_id, movie_id):
        return self.avg

class UserAverageBased(Predictor):
    """A marginally less silly predictor that always predicts the user's average rating."""

    # An example of how to save precomputed info
    def precompute(self):
        self.user_avg_rating = {}  # uid -> average prediction by uid
        for uid, group in itertools.groupby(self.data, key=operator.itemgetter(0)):
            u_pred = list(group)
            self.user_avg_rating[uid] = mean([t[2] for t in u_pred])
    
    def predict(self, user_id, movie_id):
        return self.user_avg_rating[user_id]

#TODO: implement the user-based CF algorithm
class UserBased(Predictor):
    """
    data = [(uid, movie_id, pred)]
    """
    
    def precompute(self):
        # Compute all pairs similarity between users, and find the N most similar
        # users to each
        self.user_predictions = {}  # uid -> dict: movie_id -> pred
        self.user_avg_rating = {}  # uid -> average prediction by uid
        self.user_rated = {}  # uid -> set of (movie ids rated by uid)
        self.similarities = defaultdict(list) # uid -> list of (uid2, similarity)

        # group data items by user so we can compute things
        self.data.sort()
        for uid, group in itertools.groupby(self.data, key=operator.itemgetter(0)):
            u_pred = list(group)
            self.user_avg_rating[uid] = mean([t[2] for t in u_pred])
            
            # TODO: Compute other useful info: e.g. user_predictions and user_rated

        N = self.options.N

        def similarity(u1, u2):
            # Compute the pearson correlation (notes -- Eq 16.2)
            return 0  # TODO: implement

        # TODO: compute all similarities, truncate to neighbourhood of size N
        
        logging.debug("Similarities: ")
        logging.debug(self.similarities)
        
    
    def predict(self, user_id, movie_id):
        r = self.user_avg_rating[user_id]

        # TODO: implement, using precomputed info
        
        return r

#TODO: implement the item-based CF algorithm
class ItemBased(Predictor):
    def precompute(self):
        pass
    
    def predict(self, user_id, movie_id):
        return 3.0

#TODO: implement your own predictor
class MagicBased(Predictor):
    pass

